#!/usr/bin/python3

import collections
import dataclasses
import functools
import itertools
import operator
from typing import Iterable, NamedTuple

import claripy
import itanium_demangler as demangler
import msgspec
from cle.backends.symbol import Symbol
from cle.memory import Clemory

import angr

from .. import demangler_helpers as dh

# hotfix for demangler
demangler._is_ctor_or_dtor = demangler.is_ctor_or_dtor


def read_cstring(
    mem: Clemory, addr: int, encoding: str = "utf-8", chunk: int = 16, **kwargs
) -> str:
    """unpacks a variable-length, zero-terminated string from the binary"""
    try:
        unpacker = itertools.takewhile(
            lambda x: x > 0,
            itertools.chain.from_iterable(
                mem.load(addr + o, chunk) for o in itertools.count(0, chunk)
            ),
        )
        return bytes(unpacker).decode(encoding, **kwargs)
    except Exception:
        pass
    return ""


def _reorder_vfns_windows_estimate(symbols: list[Symbol]) -> list[Symbol]:
    # reorders a given subclass-level slice of linux symbols to reflect windows ordering
    name_buckets = collections.defaultdict(list)
    for n, symbol in enumerate(symbols):
        # collect overrides into buckets based on function name
        dmsym = demangler.parse(symbol.name)
        if dmsym:
            name_buckets[dh.extract_method_fname(dmsym)].append(symbol)
        else:
            # HACK: preserves positions of references to __cxa_pure_virtual
            name_buckets[(dmsym, n)].append(symbol)

    # on windows, overloads are made consecutive and in reverse of declared order
    # iteration order is guaranteed as of py3.7+ to be the insertion order,
    # so this should output symbols otherwise in their original order
    return list(itertools.chain.from_iterable(reversed(syms) for syms in name_buckets.values()))


def _allow_symbol_for_windows_vtable(
    sym: Symbol, thunk_fns: set[tuple[demangler.Node, ...]]
) -> bool:
    dmsym = demangler.parse(sym.name)
    if dmsym:
        # MSVC only provides one dtor, so here we'll use the deleting one (D0)
        if dh.is_dtor(dmsym) and dh.get_dtor_type(dmsym) != "deleting":
            return False
        elif (
            not demangler.is_ctor_or_dtor(dmsym)
            and dh.extract_method_signature(dmsym) in thunk_fns
        ):
            # filter MI thunks
            return False
    else:
        # __cxa_pure_virtual returns None here; we still add it to the vtable slice
        pass
    return True


VTable = list[Symbol]


# some virtual functions cannot be disambiguated outright, or need to check the resolved
# state of the rest of the table - for the former we have to disambiguate them out-of-band
class BaseConstraint(msgspec.Struct, tag_field="constraint"):
    def apply(self, solver: claripy.Solver, symtable: dict[str, claripy.ast.bv.BV]) -> None:
        raise NotImplementedError("BaseConstraint.apply is abstract")


class SoftOrderConstraint(BaseConstraint, tag="soft"):
    # a constraint where symbols must be in ascending order but gaps may be present
    symbols: list[str]

    def apply(self, solver: claripy.Solver, symtable: dict[str, claripy.ast.bv.BV]) -> None:
        if not all(x in symtable for x in self.symbols):
            return
        solver.add(
            claripy.And(
                *(symtable[b] > symtable[a] for a, b in itertools.pairwise(self.symbols))
            )
        )


class ConsecutiveOrderConstraint(BaseConstraint, tag="consecutive"):
    # a constraint where symbols must have offsets one higher than the previous
    symbols: list[str]

    def apply(self, solver: claripy.Solver, symtable: dict[str, claripy.ast.bv.BV]) -> None:
        if not all(x in symtable for x in self.symbols):
            return
        solver.add(
            claripy.And(
                *(symtable[b] == symtable[a] + 1 for a, b in itertools.pairwise(self.symbols))
            )
        )


VTableConstraintDict = dict[str, list[SoftOrderConstraint | ConsecutiveOrderConstraint]]


class VtableDisambiguator(angr.Analysis):
    syms_by_addr: dict[int, set[Symbol]]

    # given a vtable symbol, returns childrens' vtable symbols in an undefined order
    subclass_map: dict[Symbol, set[Symbol]]

    # given a vtable symbol, returns the parents' vtable symbols in ascending order
    superclass_map: dict[Symbol, list[Symbol]]

    # we use this to test if a function belongs to a class with a vtable
    # this way we can prune static functions from the candidate pool
    vtable_instances: set[demangler.Node]

    vtable_constraint: VTableConstraintDict

    def __init__(self):
        self.loader = self.project.loader
        self.memory = self.loader.memory
        self.vtable_constraint = {}
        self.analyze()

    def analyze(self) -> None:
        self.syms_by_addr = collections.defaultdict(set)
        vtable_syms = set()
        self.vtable_instances = set()
        for symbol in self.loader.symbols:
            self.syms_by_addr[symbol.rebased_addr].add(symbol)
            if symbol.name.startswith("_ZTV") and not symbol.name.startswith("_ZTVZ"):
                vtable_syms.add(symbol)
                self.vtable_instances.add(
                    dh.extract_vtable_typename(demangler.parse(symbol.name))
                )

        # we associate a type with their subclasses so we can check them to disambiguate duplicates on the base
        self.subclass_map = collections.defaultdict(set)
        self.superclass_map = collections.defaultdict(list)

        for vtsym in vtable_syms:
            if vtsym.is_import:
                continue

            vt_typeinfo = self.memory.unpack_word(vtsym.rebased_addr + 0x4)
            for typeinfo_ptr, typeinfo_name in self.dump_class_parents(vt_typeinfo):
                svtsym = self.loader.find_symbol(f"_ZTV{typeinfo_name}")
                if not svtsym:
                    continue
                self.subclass_map[svtsym].add(vtsym)
                self.superclass_map[vtsym].append(svtsym)

    def dump_class_parents(self, typeinfo_ptr: int) -> Iterable[tuple[int, str]]:
        """
        Returns a list of mangled typenames in ascending order (towards base classes at the end).
        This takes a pointer to the class's typeinfo structure.
        """
        # HACK: adjust offset by -8h for angr to resolve to the correct symbol
        #       IDA indicates the offset is at sym+8h
        #       lief seems to resolve this correctly without offsetting by -8h
        if not typeinfo_ptr:
            return
        typeinfo_class_sym = self.loader.find_symbol(
            self.memory.unpack_word(typeinfo_ptr) - 0x8
        )
        if not typeinfo_class_sym:
            return

        typeinfo_name = read_cstring(self.memory, self.memory.unpack_word(typeinfo_ptr + 0x04))
        if typeinfo_class_sym.name == "_ZTVN10__cxxabiv120__si_class_type_infoE":
            # single inheritance
            nested_typeinfo_ptr = self.memory.unpack_word(typeinfo_ptr + 0x08)
            yield (typeinfo_ptr, typeinfo_name)
            yield from self.dump_class_parents(nested_typeinfo_ptr)
        elif typeinfo_class_sym.name == "_ZTVN10__cxxabiv121__vmi_class_type_infoE":
            # multiple inheritance
            nested_typeinfo_ptr = self.memory.unpack_word(typeinfo_ptr + 0x10)
            yield (typeinfo_ptr, typeinfo_name)
            yield from self.dump_class_parents(nested_typeinfo_ptr)
        elif typeinfo_class_sym.name == "_ZTVN10__cxxabiv117__class_type_infoE":
            yield (typeinfo_ptr, typeinfo_name)
        else:
            raise ValueError("unknown typeinfo class", typeinfo_class_sym)

    def resolve_ambiguous_vfn(
        self, vtidx: int, ambig_fnsyms: set[Symbol], related_vtsyms: set[Symbol]
    ) -> Symbol | None:
        """
        Resolves an ambiguous virtual function by attempting to match against symbols in related
        classes' vtables.
        """
        # ambiguous function
        # walk through vtables of subclasses to find matching prototype at the given position
        # if unsuccessful, walk through parent classes containing the vtable
        # returns the exact name of the symbol used for this function
        funcsig_set = set()
        for fnsym in ambig_fnsyms:
            funcsig_set.add(dh.extract_method_signature(demangler.parse(fnsym.name)))

        non_class_syms = set()
        for sym in ambig_fnsyms:
            fnname = dh.extract_function_name(demangler.parse(sym.name))
            if not any(vtname == fnname[: len(vtname)] for vtname in self.vtable_instances):
                non_class_syms.add(sym)

        if len(ambig_fnsyms - non_class_syms) == 1:
            # all but one of the ambiguous symbols are not potential virtual class methods
            return (ambig_fnsyms - non_class_syms).pop()

        if len(funcsig_set) == 1:
            # we only have one name-signature combination here
            return set(ambig_fnsyms).pop()

        rebased_fnsym_addrs = set(fnsym.rebased_addr for fnsym in ambig_fnsyms)

        # if all the possible symbols are destructors, return based on priority
        # this is technically unambiguous since GCC will always alias base/complete, but for the
        # sake of this analysis there must be only one symbol for each entry
        if all(dh.is_dtor(demangler.parse(sym.name)) for sym in ambig_fnsyms):
            for dtype, sym in itertools.product(("complete", "base"), ambig_fnsyms):
                if dh.get_dtor_type(demangler.parse(sym.name)) == dtype:
                    return sym

        for fnsym in ambig_fnsyms - non_class_syms:
            funcsig = dh.extract_method_signature(demangler.parse(fnsym.name))
            for svt in related_vtsyms:
                # only check functions at the same index
                svt_first, *_ = self.get_vfptrs_from_table(svt)
                if vtidx >= len(svt_first):
                    continue
                subfn_addr = svt_first[vtidx]

                if subfn_addr in rebased_fnsym_addrs:
                    # skip vtable entries that call the same exact function
                    continue

                subfns = self.syms_by_addr.get(subfn_addr)
                if not subfns:
                    continue

                for subfn in subfns:
                    sub_funcsig = dh.extract_method_signature(demangler.parse(subfn.name))
                    if funcsig != sub_funcsig:
                        continue

                    # we found a subclass with a name / prototype that matches the base class
                    return fnsym
        return None

    def get_possible_vtable_set_candidates(self, vtsym: Symbol, vtidx: int) -> set[Symbol]:
        # It's possible that this class inherited a base version of a method that is specialized
        # in a different part of the class hierarchy, so find the base implementor then return
        # it along with its descendants.
        psym = self.get_parent_vt_implementing(vtsym, vtidx)
        return {psym} | self.subclass_map[psym]

    def get_parent_vt_implementing(self, vtsym: Symbol, vtidx: int) -> Symbol:
        # walk up parents until we reach the first one to implement this method
        for super_vtsym, gp_vtsym in itertools.pairwise([vtsym, *self.superclass_map[vtsym]]):
            vt_first, *_ = self.get_vfptrs_from_table(gp_vtsym)
            if len(vt_first) < vtidx:
                return super_vtsym
        return self.superclass_map[vtsym][-1]

    @functools.cache
    def get_vfptrs_from_table(self, vtsym: Symbol) -> list[list[int]]:
        # returns a list of addresses for each vtable present on the class
        class VTableFunction(NamedTuple):
            tblidx: int
            address: int

        table_index = 0
        function_list: list[VTableFunction] = []

        # we include the initial upcast offset and class typeinfo here
        vtable_range = enumerate(range(vtsym.rebased_addr, vtsym.rebased_addr + vtsym.size, 4))

        for n, addr in vtable_range:
            # get symbols that map to that address
            # NOTE: classes that don't implement virtual destructors have nullptrs
            deref = self.loader.fast_memory_load_pointer(addr)
            fnsyms = set(self.syms_by_addr.get(deref) or set()) if deref else set()
            if not fnsyms:
                # vtable boundary; consume typeinfo so it doesn't get added to the list
                # NOTE: we don't actually care if the indices skip; all that matters is that the
                # functions are grouped correctly
                table_index += 1
                next(vtable_range)

                for n1, addr1 in vtable_range:
                    # consume any leading nullptrs (destructors that were optimized out)
                    # this also consumes the next actual function pointer
                    # we could probably get away with looking for it in the parent class...
                    deref = self.loader.fast_memory_load_pointer(addr1)
                    function_list.append(VTableFunction(table_index, deref))
                    if deref != 0:
                        break
                continue
            function_list.append(VTableFunction(table_index, deref))

        return [
            list(vfn.address for vfn in vtbl)
            for tblidx, vtbl in itertools.groupby(
                function_list, key=operator.attrgetter("tblidx")
            )
        ]

    @functools.cache
    def get_windows_vtables_from(self, vt: Symbol) -> VTable:
        vt_parent_spans = [0]
        for vt_parent in reversed(self.superclass_map[vt]):
            vt_parent_first, *_ = self.get_vfptrs_from_table(vt_parent)

            # it's possible for a parent table to have more functions than its child
            # e.g. CAutoGameSystemPerFrame has more functions present than CGameRules
            span = len(vt_parent_first)
            if span > vt_parent_spans[-1]:
                vt_parent_spans.append(span)

        vt_first, *vt_others = self.get_vtables_from_address(vt)
        thunk_fns = set()
        for vt_other in vt_others:
            for sym in vt_other:
                dmsym = demangler.parse(sym.name)
                if dmsym and dmsym.kind == "nonvirt_thunk":
                    thunk_fns.add(dh.extract_method_signature(dmsym))

        vt_out = []
        for vt_low, vt_high in itertools.pairwise(vt_parent_spans):
            # we can only reorder overloads within the class they were initially specified
            # e.g. CTFPlayer's ChangeTeam cannot be merged with CBaseEntity's
            class_vfns = []

            for sym in vt_first[vt_low:vt_high]:
                if _allow_symbol_for_windows_vtable(sym, thunk_fns):
                    class_vfns.append(sym)
            vt_out.extend(_reorder_vfns_windows_estimate(class_vfns))

        return vt_out

    def get_vfn_from_parent(self, vt: Symbol, vtidx: int) -> int:
        # HACK: some virtual destructors got optimized out and are represented by nullptrs
        #       in that case we try to match functions from the parent
        for parent_vt in self.superclass_map[vt]:
            pvptr_first, *_ = self.get_vfptrs_from_table(parent_vt)
            if vtidx > len(pvptr_first):
                break
            vptr = pvptr_first[vtidx]
            if vptr:
                return vptr
        return 0

    @functools.cache
    def get_vtables_from_address(self, vt: Symbol) -> list[VTable]:
        # returns a list of vtables for each vtable present on the class
        @dataclasses.dataclass
        class VTableFunction:
            tblidx: int
            possible_syms: set[Symbol]

        disambiguated_func_sigs = set()
        function_list: list[VTableFunction] = []
        vptr_lists = self.get_vfptrs_from_table(vt)
        for table_index, vptrs in enumerate(vptr_lists):
            for n, vptr in enumerate(vptrs):
                # get symbols that map to that address
                vptr = vptr or self.get_vfn_from_parent(vt, n)
                if vptr == 0:
                    call_unexpected = self.loader.find_symbol("__cxa_call_unexpected")
                    assert call_unexpected
                    function_list.append(VTableFunction(table_index, {call_unexpected}))
                    continue

                fnsyms = set(self.syms_by_addr.get(vptr) or set())

                if len(fnsyms) > 1:
                    # function in vtable is referenced by multiple names; perform disambiguation
                    matched_overload = self.resolve_ambiguous_vfn(
                        n, fnsyms, self.get_possible_vtable_set_candidates(vt, n)
                    )

                    # it's possible that the other function(s) is/are resolveable.

                    if matched_overload:
                        # within a vtable we expect a non-extern method signature
                        # (name + params) to resolve exactly once, so as a result we can
                        # eliminate a function and its other class variants from candidacy
                        # elsewhere
                        disambiguated_func_sigs.add(
                            dh.extract_method_signature(demangler.parse(matched_overload.name))
                        )
                        function_list.append(VTableFunction(table_index, {matched_overload}))
                        continue

                function_list.append(VTableFunction(table_index, fnsyms))

        for vfn in function_list:
            # disambiguated_func_sigs may include an already unique match so we skip those
            if len(vfn.possible_syms) > 1:
                disambiguated_functions = set(
                    x
                    for x in vfn.possible_syms
                    if dh.extract_method_signature(demangler.parse(x.name))
                    in disambiguated_func_sigs
                )
                vfn.possible_syms -= disambiguated_functions

        if self.vtable_constraint:
            solver = claripy.Solver()
            solver_asts = {}
            positional_constraints = collections.defaultdict(set)
            for n, vfn in enumerate(function_list):
                if len(vfn.possible_syms) == 1:
                    sym = set(vfn.possible_syms).pop()
                    # symbolic names for constant values - we don't use these ourselves; they're here
                    # so user-defined tooling can specify them relative to variables as needed
                    solver_asts[sym.name] = claripy.BVV(n, 10)
                    continue

                for sym in vfn.possible_syms:
                    # constrain variable values to those defined as ambiguous
                    positional_constraints[sym].add(n)

            for sym, positions in positional_constraints.items():
                if sym.name not in solver_asts:
                    solver_asts[sym.name] = claripy.BVS(sym.name, 10)
                solver.add(claripy.Or(*(solver_asts[sym.name] == n for n in positions)))

            for pvt in self.superclass_map[vt]:
                constraints = self.vtable_constraint.get(pvt.name)
                if not constraints:
                    continue
                for constraint in constraints:
                    constraint.apply(solver, solver_asts)

            solutions = solver.batch_eval(solver_asts.values(), 2)
            if len(solutions) == 1:
                unique_solution = {v: k for k, v in zip(solver_asts.keys(), solutions[0])}
                for n, vfn in enumerate(function_list):
                    if (
                        len(vfn.possible_syms) > 1
                        and unique_solution
                        and n in unique_solution.keys()
                    ):
                        resolved_syms = {
                            sym
                            for sym in iter(vfn.possible_syms)
                            if sym.name == unique_solution[n]
                        }
                        vfn.possible_syms = resolved_syms

        # final uniqueness pass
        # raise exception if we still have multiple possible symbols for a given position
        for n, vfn in enumerate(function_list):
            if len(vfn.possible_syms) == 1:
                continue

            # we should never receive an empty ``VTableFunction.possible_syms``
            vt_name = demangler.parse(vt.name)
            candidate_names = set(sym.name for sym in vfn.possible_syms)
            raise Exception(
                f"Ambiguity in {vt_name} position {n}; candidates {candidate_names}"
            )

        return [
            list(vfn.possible_syms.pop() for vfn in vtbl)
            for tblidx, vtbl in itertools.groupby(
                function_list, key=operator.attrgetter("tblidx")
            )
        ]


angr.analyses.AnalysesHub.register_default("VtableDisambiguator", VtableDisambiguator)
